/*****************************************************************
 * Copyright (c) ParaSoft Corporation 1991.  All rights reserved.*
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF ParaSoft       *
 * The copyright notice above does not evidence any              *
 * actual or intended publication of such source code.           *
 *****************************************************************/

#include "extern.h"

#define LABEL_INIT	7000  /* starting index for generated labels */

#define LABEL_DEF		1
#define LABEL_REF		2
#define LABEL_FORMAT	3
#define LABEL_IO		4
#define LABEL_TEMP		5

static int labelnum = LABEL_INIT;

/************************** CHECKING PART **********************************/

/*
1) checks that the labels are defined only once, and that the referenced labels are
   defined, and that form of label is OK.

We could find a label in 4 cases:
1) ContinueStmt ==> IS_label_def
2) GotoStmt ... ==> IS_label_ref
3) FormatStmt   ==> IS_label_format
4) I/O Stmt     ==> IS_label_io
*/

static label_set_type (entry)
T_Entry *entry;
{
	if (entry->typ != (Typ*) 0)
		return;

	entry->typ = typ_malloc ();
	entry->typ->etype = Etype_INTEGER_0;
}

static char *label_Uname (name) /* checks that form of label is Ok, and makes it unique */
char *name;
{
	int i, non0, len;
	char *Uname;

	if ((len = strlen (name)) > 5) {
		sprintf (serr, "label '%s': label can at most have 5 digits", name);
		fatal (serr);  /* E_3_4_5 */
	}

	non0 = 0;
	Uname = U_str_alloc (len);
	for (i = 0; i < len; i++)
		if (non0 != 0 || name[i] != '0')
			Uname [non0++] = name [i];

	if (non0 == 0) {
		sprintf (serr, "label '%s': at least one digit in a label must be nonzero", name);
		fatal (serr);  /* E_3_4_6 */
		Uname = "0";
	}

	return Uname;
}

static T_Entry *label_install (label, what)
Node *label;
int what;
{
	char *name = label->com.str;
	char *Uname = label_Uname (name);
	T_Entry *entry;

	if ((entry = lookup_loc (Uname)) != TN) {
		if ((what == LABEL_DEF || what == LABEL_FORMAT)
		&& (IS_label_def (entry) || IS_label_format (entry))) {
			sprintf (serr, "label '%s' already defined", name);
			fatal (serr);
		}
	}
	else {
		entry = install_loc (name, TAB_LABEL, label);
		label_set_type (entry);
	}

	switch (what) {

	case LABEL_DEF:
		if (IS_label_io (entry)) {
/*!!! AssignStmt screws up
			sprintf (serr, "label '%s' must correspond to a format statement", name);
			fatal (serr);
*/
			;
		}
		else {
			SET_label_def (entry);
		}
		break;

	case LABEL_REF:
		if (IS_label_format (entry)) {
			sprintf (serr, "label '%s' belongs to a format statement", name);
			fatal (serr);
		}
		else {
			SET_label_ref (entry);
		}
		break;

	case LABEL_FORMAT:
		if (IS_label_ref (entry)) {
/*!!! AssignStmt screws up
			sprintf (serr, "label '%s' must correspond to an executable statement", name);
			fatal (serr);
*/
			;
		}
		else {
			SET_label_format (entry);
		}
		break;

	case LABEL_IO:
		if (IS_label_def (entry)) {
			sprintf (serr, "label '%s' belongs to an executable statement", name);
			fatal (serr);
		}
		else {
			SET_label_io (entry);
		}
		break;

	case LABEL_TEMP:
		SET_label_temp (entry);
		break;

#ifdef CHECK
	default:
		SYSERR;
#endif
	}

	return entry;
}

T_Entry *label_install_def (label)
Node *label;
{
	return label_install (label, LABEL_DEF);
}

T_Entry *label_install_ref (label)
Node *label;
{
	return label_install (label, LABEL_REF);
}

T_Entry *label_install_format (label)
Node *label;
{
	return label_install (label, LABEL_FORMAT);
}

T_Entry *label_install_io (label)
Node *label;
{
	return label_install (label, LABEL_IO);
}

label_end ()  /* checks that all referenced labels are defined */
{
	T_Entry *entry;

	for_TABLE_LOOP (entry) {
		if (!IS_label (entry))
			continue;

		if (!IS_label_def (entry) && !IS_label_format (entry))
			if (IS_label_ref (entry)) {
/*!!! AssignStmt screws up
				sprintf (serr, "label '%s' with executable statement not defined", entry->name);
				fatal (serr);
*/
			}
			else if (!IS_label_temp (entry)) {
				sprintf (serr, "format with label '%s' not defined", entry->name);
				fatal (serr);
			}

		if (!IS_label_ref (entry) && !IS_label_io (entry))
		 	if (IS_label_def (entry)) {
				sprintf (serr, "executable statement with label '%s' not referenced", entry->name);
				/*!!! gives unwanted warnings for labels not detected (i.e. end = 100)
				warning (serr);
				*/
			}
			else {
				; /*!!! for when we do I/O
				sprintf (serr, "format with label '%s' not referenced", entry->name);
				warning (serr);
				*/
			}
	}

/* TRANSLATION */

	labelnum = LABEL_INIT;
}

/************************* TRANSLATION PART *********************************/

Node *label_make ()
{
	char label [10], *tempname;
	Node *Label;
	T_Entry *entry;

	sprintf (label, "#%d", labelnum++);

	Label = make (INTEGER_CONSTANT);
	Label->com.typ = typ_malloc ();
	Label->com.typ->etype = Etype_INTEGER_0;
	Label->com.str = U_str_copy (label);

	entry = label_install (Label, LABEL_TEMP);

	tempname = name_label (entry->orig_name);

	while (lookup_all (tempname) != TN) {
		U_free (tempname);
		tempname = name_label_num ();
	}

	entry->tempname = tempname;

	return Label;
}
